// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/16AISS16AO2/utils/util_fsamp_ai.c $
// $Rev: 52934 $
// $Date: 2023-04-20 15:50:26 -0500 (Thu, 20 Apr 2023) $

// 16AISS16AO2: Utilities: source file

#include "main.h"



// macros *********************************************************************

#define	ABS(v)					(((v) < 0) ? (-(v)) : (v))
#define	LIMIT_MAX(v,max)		(((v) > (max)) ? (max) : (v))
#define	LIMIT_MIN(v,min)		(((v) < (min)) ? (min) : (v))
#define	LIMIT_RANGE(v,min,max)	(LIMIT_MIN((LIMIT_MAX((v),(max))),(min)))



/******************************************************************************
*
*	Function:	aiss16ao2_fsamp_ai_compute
*
*	Purpose:
*
*		Determine the best configuration for the specified input sample rate.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	Work verbosely?
*
*		fsamp	The desired sample rate.
*
*		ndiv	The Ndiv value is put here. MUST NOT BE NULL!
*
*		range	The range value is put here. MUST NOT BE NULL!
*
*		rate	Report the sample rate achieved here, if non-NULL.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int aiss16ao2_fsamp_ai_compute(int fd, int index, int verbose, s32 fsamp, s32* ndiv, s32* range, double* rate)
{
	double	delta;
	double	delta_best	= FLT_MAX;
	int		errs		= 0;
	double	fsamp_best	= 0;
	double	fsamp_got;
	s32		fsamp_max;	// Queried value.
	s32		fsamp_min;	// Queried value.
	double	fsamp_want;
	int		init		= 0;	// Have we performed the first run initialization?
	s32		master;
	s32		nd;
	s32		nd_max;
	s32		nd_min;
	s32		nd_use;
	s32		ndiv_best	= 0;
	s32		ndiv_max;	// Queried value.
	s32		ndiv_min;	// Queried value.

	if (verbose)
	{
		gsc_label_index("AI Sample Rate", index);
		gsc_label_long_comma(fsamp);
		printf(" S/S");
	}

	errs	+= aiss16ao2_query(fd, -1, 0, AISS16AO2_QUERY_FSAMP_MAX, &fsamp_max);
	errs	+= aiss16ao2_query(fd, -1, 0, AISS16AO2_QUERY_FSAMP_MIN, &fsamp_min);
	errs	+= aiss16ao2_query(fd, -1, 0, AISS16AO2_QUERY_MASTER_CLOCK, &master);
	errs	+= aiss16ao2_query(fd, -1, 0, AISS16AO2_QUERY_NDIV_MAX, &ndiv_max);
	errs	+= aiss16ao2_query(fd, -1, 0, AISS16AO2_QUERY_NDIV_MIN, &ndiv_min);

	fsamp_want	= (double) LIMIT_RANGE(fsamp, fsamp_min, fsamp_max);
	nd			= master / fsamp_want;
	nd_min		= nd - 1;
	nd_max		= nd + 1;
	nd_min		= (nd_min < ndiv_min) ? ndiv_min : nd_min;
	nd_max		= (nd_max > ndiv_max) ? ndiv_max : nd_max;
	ndiv[0]		= nd;

	for (nd = nd_min; nd <= nd_max; nd++)
	{
		if (nd < nd_min)
			nd_use	= nd_min;
		else if (nd > nd_max)
			nd_use	= nd_max;
		else
			nd_use	= nd;

		fsamp_got	= ((double) master) / nd_use;
		delta		= fsamp_got - fsamp_want;

		if ((init == 0) || ((ABS(delta)) < (ABS(delta_best))))
		{
			init		= 1;
			delta_best	= delta;
			fsamp_best	= fsamp_got;
			ndiv_best	= nd_use;
		}
	}

	ndiv[0]		= ndiv_best;

	if (ndiv[0] >= 800000L)
		range[0]	= AISS16AO2_FSAMP_RANGE_LOW;
	else
		range[0]	= AISS16AO2_FSAMP_RANGE_HIGH;

	if (rate)
		rate[0]		= fsamp_best;

	if (verbose)
	{
		printf("%s  (", errs ? " FAIL <---" : "");
		printf("Fref ");
		gsc_label_long_comma((long) master);
		printf(", Ndiv ");
		gsc_label_long_comma((long) ndiv_best);
		printf(", %s Range", range[0] ? "High" : "Low");
		printf(", Fsamp %.3f", fsamp_best);
		printf(")\n");
	}

	return(errs);
}



/******************************************************************************
*
*	Function:	aiss16ao2_fsamp_ai_get
*
*	Purpose:
*
*		Determine the input sample rate.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		verbose	Work verbosely?
*
*		get		Store the sample rate here, if non-NULL.
*
*		range	Store the range setting here, if non-NULL.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int aiss16ao2_fsamp_ai_get(int fd, int verbose, s32* get, s32* range)
{
	s32	mode;
	s32	ndiv		= -1;
	s32	enable		= -1;
	int	errs		= 0;
	s32	master;
	s32	set			= 0;

	errs	+= aiss16ao2_ai_clock_mode	(fd, -1, 0, -1, &mode);
	errs	+= aiss16ao2_gen_a_enable	(fd, -1, 0, -1, &enable);
	errs	+= aiss16ao2_gen_a_ndiv		(fd, -1, 0, -1, &ndiv);
	errs	+= aiss16ao2_query			(fd, -1, 0, AISS16AO2_QUERY_MASTER_CLOCK, &master);

	if (errs)
	{
		set	= 0;

		if (verbose)
			printf("FAIL <---  (%d. service errors)\n", __LINE__);
	}
	else if (mode == AISS16AO2_CLOCK_MODE_IN)
	{
		set	= 0;

		if (verbose)
			printf("Unknown  (Clock Mode = Input)\n");
	}
	else if (enable == AISS16AO2_GEN_ENABLE_NO)
	{
		set	= 0;

		if (verbose)
			printf("0 S/S  (Rate Gen. A is disabled.)\n");
	}
	else
	{
		set	= master / ndiv;

		if (verbose)
		{
			gsc_label_long_comma(set);
			printf(" S/S  (Source = Rate Gen A, Ref = ");
			gsc_label_long_comma(master);
			printf(", Ndiv = ");
			gsc_label_long_comma(ndiv);
			printf(", Range %s)\n", range[0] ? "High" : "Low");
		}
	}

	if (get)
		get[0]	= set;

	return(errs);
}



/******************************************************************************
*
*	Function:	aiss16ao2_fsamp_ai_report
*
*	Purpose:
*
*		Determine and report the input sample rate for the specified channel.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	Work verbosely?
*
*		chan	The index of the channel of interest.
*
*		get		Store the sample rate here, if non-NULL.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int aiss16ao2_fsamp_ai_report(int fd, int index, int verbose, int chan, s32* get)
{
	s32		bit			= 0x1L << chan;
	char	buf[64];
	int		errs;
	s32		select;
	s32		set			= 0;
	s32		range		= 0;

	if (verbose)
	{
		sprintf(buf, "AI Channel %d", chan);
		gsc_label_index(buf, index);
	}

	errs	= aiss16ao2_ai_chan_sel(fd, -1, 0, -1, &select);

	if (select & bit)
	{
		// The channel is enabled.
		errs	+= aiss16ao2_fsamp_ai_get(fd, verbose, &set, &range);
	}
	else
	{
		// The channel is disabled.
		set	= 0;

		if (verbose)
			printf("0 S/S  (Channel Disabled)\n");
	}

	if (get)
		get[0]	= set;

	return(errs);
}



/******************************************************************************
*
*	Function:	aiss16ao2_fsamp_ai_report_all
*
*	Purpose:
*
*		Determine and report the input sample rate for all channels.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	Work verbosely?
*
*		fsamp	Report the overall sample rate here.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int aiss16ao2_fsamp_ai_report_all(int fd, int index, int verbose, s32* fsamp)
{
	s32	chans;
	int	chan;
	int	errs;
	s32	sps;
	s32	total	= 0;

	if (verbose)
		gsc_label_index("AI Sample Rates", index);

	errs	= aiss16ao2_query(fd, -1, 0, AISS16AO2_QUERY_CHANNEL_AI_QTY, &chans);

	if (errs)
	{
	}
	else
	{
		if (verbose)
		{
			printf("\n");
			gsc_label_level_inc();
		}

		for (chan = 0; chan < chans; chan++)
		{
			sps		= 0;
			errs	+= aiss16ao2_fsamp_ai_report(fd, -1, verbose, chan, &sps);
			total	+= sps;
		}

		if (verbose)
		{
			gsc_label("Overall AI Rate");
			gsc_label_long_comma(total);
			printf(" S/S\n");
			gsc_label_level_dec();
		}
	}

	if (fsamp)
		fsamp[0]	= total;

	return(errs);
}


